from peak.events import trellis
from peak.util import addons, decorators
from urllib import unquote

__all__ = [
    'Switch', 'EventHandler', 'Child', 'Channel', #'Conference', 'Member',
    'parse_headers',
]

def parse_headers(text, encoded=False):
    if '\n\n' in text:
        head, body = text.split('\n\n',1)
    else:
        head, body = text, None
    d = {}
    for line in head.splitlines():
        k, v = line.split(': ',1)
        if encoded and '%' in v:
            v = unquote(v)
        d[k] = v
    if body:
        d['body'] = body
    return d


















def make(factory):
    return lambda self, factory=factory: factory()

class Pipe(trellis.Component):
    """Allow one or more writers to send data to zero or more readers"""

    output = trellis.todo(make(list))
    input  = output.future

    decorators.decorate(trellis.modifier)
    def append(self, value):
        self.input.append(value)

    decorators.decorate(trellis.modifier)
    def extend(self, sequence):
        self.input.extend(sequence)

    def __iter__(self):
        return iter(self.output)

    def __contains__(self, value):
        return value in self.output

    def __len__(self):
        return len(self.output)

    def __repr__(self):
        return repr(self.output)













class EventHandler(trellis.Component):
    """Something that may or may not exist, according to events"""

    events = trellis.rule(make(Pipe))
    closed = trellis.value(False)

    event_attributes = ()

    @trellis.rule
    def _update(self):
        for event in self.events:
            for attrname, event_key in self.event_attributes:
                if event_key in event:
                    setattr(self, attrname, event[event_key])                    
            getattr(self, event['Event-Name'], self.ANY_EVENT)(event)

    def ANY_EVENT(self, event):
        pass

    def close(self, reason=True):
        self.closed = reason




















class Switch(EventHandler):
    """Process inbound messages from FreeSWITCH"""

    trellis.rules(
        messages  = make(Pipe),     # misc. content-types
        transmit  = make(Pipe),     # strings to send back to switch
        errors    = make(Pipe),     # errors to log or whatever
        log_data  = make(Pipe),     # incoming log data
        message_handlers  = lambda self: {
            'text/event-plain':
                lambda self,t,h,b: self.events.append(parse_headers(b, True)),
            'command/reply':
                lambda self,t,h,b: self.callback(h['Reply-Text']),
            'api/response':
                lambda self,t,h,b: self.callback(b),
            'auth/request':
                lambda self,t,h,b: setattr(self, 'auth_requested', True),
            'log/data':  lambda self,t,h,b: self.log_data.append(b),
        },
        background_jobs = make(dict),
        commands_sent = make(list),
        watch_events = lambda self: trellis.Set([]),
        channels = make(trellis.Set),
        #conferences = make(trellis.Set),
    )

    credential = 'ClueCon'  # default password
    auth_requested = trellis.receiver(False)
    is_ready = trellis.value(False)
    log_level = trellis.value(None)
    _buffer = ''
    _bytes_wanted = 0

    def ANY_EVENT(self, event):
        if 'Unique-ID' in event:
            Channel(self, event['Unique-ID']).events.append(event)
    #    if 'Conference-Name' in event:
    #        Conference(self, event['Conference-Name']).events.append(event)



    @trellis.modifier
    def receivedBytes(self, bytes):
        """Turn incoming bytes into calls to ``self.receivedMessage()``"""
        data = self._buffer + bytes
        bytes_wanted = self._bytes_wanted

        while len(data) >= bytes_wanted:

            if not bytes_wanted:
                if '\n\n' in data:
                    headers, data = data.split('\n\n', 1)
                    msg = self._message = parse_headers(headers)
                    if 'Content-Length' in msg:
                        bytes_wanted = int(msg.pop('Content-Length'))
                    else:
                        bytes_wanted = 0
                else:
                    break

            if len(data) >= bytes_wanted:
                body = data[:bytes_wanted]
                data = data[bytes_wanted:]
                msg = self._message
                self.receivedMessage(msg.pop('Content-Type'), msg, body)
                bytes_wanted = 0

        self._buffer = data
        self._bytes_wanted = bytes_wanted

    def receivedMessage(self, msgtype, headers=None, body=''):
        """Process a received message"""
        self.message_handlers.get(msgtype, self._default_receiver)(
            self, msgtype, headers or {}, body
        )

    def _default_receiver(self, _self, msgtype, headers, body):
        self.messages.append((msgtype, headers, body))




    @trellis.modifier
    def command(self, cmd, callback=None, *headers):
        """Send a command, invoking an optional `callback` when done"""
        cmd = '%s\n%s\n' % (
            cmd, ''.join('%s: %s\n' % pair for pair in headers)
        )
        self.transmit.append(cmd)
        self.commands_sent.append(callback)
        trellis.on_undo(self.commands_sent.pop)

    def api(self, cmd, callback, *headers):
        """Set up a background API job and invoke `callback` when done"""
        def queue_job(reply):
            if reply.startswith('+OK Job-UUID: '):
                uuid = reply.split(': ',1)[1]
                self.background_jobs[uuid] = callback
                trellis.on_undo(self.background_jobs.pop, uuid, None)
            else:
                callback(reply)
        self.command('bgapi '+cmd, queue_job, *headers)

    @trellis.modifier
    def callback(self, reply):
        if self.commands_sent:
            cb = self.commands_sent.pop(0)
            trellis.on_undo(self.commands_sent.insert, 0, cb)
            if cb is not None: cb(reply)

    @trellis.modifier
    def BACKGROUND_JOB(self, event):
        jobs = self.background_jobs
        uuid = event['Job-UUID']
        callback = jobs.pop(uuid, None)
        if callback is not None:
            trellis.on_undo(jobs.__setitem__, uuid, callback)
            callback(event['body'])





    @trellis.rule
    def _set_event_types(self):
        if self.is_ready:
            events = list(self.watch_events)
            if (
                'all' not in events and 'ALL' not in events
                and 'BACKGROUND_JOB' not in events
            ):
                events.append('BACKGROUND_JOB')
            self.command('noevents')
            self.command('event plain '+' '.join(events))

    @trellis.rule
    def _send_authorization(self):
        if self.auth_requested:
            self.is_ready = False
            self.command('auth '+self.credential, self._receive_authorization)

    def _receive_authorization(self, response):
        if response.startswith('+OK'):
            self.is_ready = True
        else:
            self.errors.append(response)

    @trellis.rule
    def _set_log_level(self):
        if self.is_ready:
            if self.log_level is not None:
                self.command('log '+str(self.log_level))
            else:
                self.command('nolog')

    def close(self, reason=True):
    #    for conf in list(self.conferences):
    #        conf.close(reason)
        for channel in list(self.channels):
            channel.close(reason)
        super(Switch, self).close(reason)



    @classmethod
    def client_factory(cls, callback, **kw):
        from freeswytch.twisted_support import SwitchFactory
        return SwitchFactory(cls, kw, callback)

    @classmethod
    def connect(cls, host='127.0.0.1', port=8021, timeout=30, bindAddress=None,
        callback=None, **kw
    ):
        if callback is None:
            from twisted.internet import defer
            d = defer.Deferred()
            callback = d.callback
        else:
            d = None
        factory = cls.client_factory(callback, **kw)
        from twisted.internet import reactor
        connect = reactor.connectTCP(host, port, factory, timeout, bindAddress)
        if d is not None:
            return d
        return connect


    #@classmethod
    #def server_factory(cls, callback, **kw):
    #    return cls.client_factory(callback, is_ready=True, **kw)

    #@classmethod
    #def listen(cls, port, callback, backlog=50, interface='', **kw):
    #    factory = cls.server_factory(callback, **kw)
    #    from twisted.internet import reactor
    #    return reactor.listenTCP(port, factory, backlog, interface)









class Collaborator(addons.AddOn):
    """Keyed Add-on with owner/collection support"""

    owner      = None
    key        = None
    owning_set = None 

    def __init__(self, owner, key):
        self.owner = owner
        self.key = key

    @classmethod
    def __class_call__(cls, owner, key):
        self = super(Collaborator, cls).__class_call__(owner, key)
        if self.owning_set:
            siblings = getattr(owner, self.owning_set)
            if self not in siblings:
                siblings.add(self)
        return self

    @classmethod
    def delete_from(cls, ob, *key):
        """Ensure an aspect of this type for the given key does not exist"""
        if cls.exists_for(ob, *key):
            self = cls(ob, *key)
            if self.owning_set:
                siblings = getattr(self.owner, self.owning_set)
                if self in siblings:
                    siblings.remove(self)
            super(Collaborator, self).delete_from(ob, *key)

    def detach(self):
        self.delete_from(self.owner, self.key)








class Child(Collaborator, EventHandler):
    """Event handler that belongs to a parent object"""

    def __init__(self, owner, key, **args):
        Collaborator.__init__(self, owner, key)
        EventHandler.__init__(self, **args)

    @trellis.modifier
    def close(self, reason=True):
        self.detach()
        super(Child, self).close(reason)

    @trellis.modifier
    def detach(self):
        super(Child, self).detach()
        trellis.on_undo(self._reattach)
        
    @classmethod
    @trellis.modifier
    def __class_call__(cls, owner, key):
        # This method *must* be a modifier, otherwise there will be a recalc
        # cycle between the parent set for children of this type, and the
        # EventHandler._update rule.
        return super(Child, cls).__class_call__(owner, key)

    def _reattach(self):
        addons.addons_for(self.owner)[self.addon_key(self.key)] = self
        



        









class Channel(Child):
    """A single active channel"""

    owning_set = 'channels'

    trellis.values(
        user = None, name = None, number = None,
        state = None, destination = None
    )
    event_attributes = dict(
        user   = 'Caller-Username',
        name   = 'Caller-Caller-ID-Name',
        number = 'Caller-Caller-ID-Number',
        state  = 'Channel-State',
        destination = 'Caller-Destination-Number',
    ).items()
    
    def hangup(self, cause='NORMAL_CLEARING', callback=None):
        """Disconnect this channel"""
        self.command('hangup', callback, ('hangup-cause', cause))

    def execute(self, app, data=None, times=1, callback=None):
        """Execute a standard app+data command, `times` times"""
        headers = [('execute-app-name', app)]
        if data is not None:
            headers.append(('execute-app-arg', data))
        if times != 1:
            headers.append(('loops', str(times)))
        self.command('execute', callback, *headers)

    def command(self, cmd, callback=None, *headers):
        """Send a channel command (e.g. 'hangup', 'nomedia', 'execute')"""
        self.owner.command(
            'SendMsg '+self.key, callback, ('call-command', cmd), *headers
        )

    def CHANNEL_DESTROY(self, event):
        self.close(event.get('variable_hangup_cause', 'NORMAL_CLEARING'))



    def __repr__(self):
        return "Channel(key=%r, user=%r, name=%r, number=%r)" % (
            self.key, self.user, self.name, self.number)





































